'Filename   NWT-VFO_Mega168_102.bas
'Version    1.02
'Date       2022-02-+19
'Purpose    Control of FA-NWT VFO
'Author     Georg Latzel, DL6GL
'Copyright  (c) Georg Latzel, free for private use under GNU license
'Hardware   ATmega168 (TQFP)
'           Crystal 14.7456 MHz for good UART performance
'           Encoder, 3 keys (left, right, mode), 16x2 LCD
'           Encoder A        @ PINC.0
'           Encoder B        @ PINC.1
'           Encoder key      @ PINC.2
'           Key mode         @ PINC.3
'           Key cursor right @ PINC.4
'           Key cursor left  @ PINC.5
'           LCD   RS         @ PD5
'           LCD   E          @ PD6
'           LCD   DB4        @ PD7
'           LCD   DB5        @ PB0
'           LCD   DB6        @ PB1
'           LCD   DB7        @ PB2
'           Communication to NWT-PIC:
'           57,600 baud, 8 bits, no parity, 1 stop bit (8N1)
'Compiler   BASCOM-AVR 2.0.8.3 (V1.02)
'           For readable "camel case" variables mark "Don't change case"
'           in BASCOM Options-Programmer-Environment-Editor

'Edit history
'V1.00      2014-01-25, first release
'V1.01      2014-01-31, Improved encoder timing (dis/enable Timer0)
'                       Encoder with DoubleStep = 0, not = 1
'                       Bull shit removed from Sub SetValue
'                       Flash used: 9,892 bytes (60%)
'V1.02      2022-02-19  Timer0 interrupt frequency increased from 72 to 120Hz
'                       for better encoder performance
'                       Flash used: 9,872 bytes (60%)
'
'==============================================================================

'ATmega168 FUSE-Bits (AVR Studio 4.0-Settings, [ ]=no check, [x]=check)
'In ATMEGA-Data sheets: Bit = 1 -> not set
'                       Bit = 0 -> set
'BODLEVEL   Brown-out detection at VCC=4.3 V
'RSTDISBL   [ ]
'DWEN       [ ]
'SPIEN      [x] Do not change!
'WDTON      [ ]
'EESAVE     [x]
'BOOTSZ     Boot Flash size=256 words start address $3F00
'BOOTRST    [ ]
'CKDIV8     [ ]
'CKOUT      [ ]
'SUT_CKSEL  Ext. Crystal Osc. 8.0- MHz, Start-up time 16k + 14ms

'Extended   0xFC
'High       0xD7
'Low        0xFF

$regfile = "m168def.dat"                          'ATMega168
$crystal = 14745600                               '14.7456 MHz xtal
$hwstack = 64
$swstack = 64
$framesize = 64
$Baud = 57600                                     'NWT PIC baud rate
Config Com1 = Dummy , Synchrone = 0 , Parity = None , Stopbits = 1 , Databits = 8 , Clockpol = 0

'Configure encoder and keys on Port A (Mega168) -------------------------------
DDRC = &B00000000                                 'PortC = input
PORTC = &B00111111                                'Pullup PC0...PC5
Key_port Alias PinC                               'Input-Port Keys
bitEnc_b Alias PINC.0                             'Encoder A        @ PINC.0
bitEnc_a Alias PINC.1                             'Encoder B        @ PINC.1
Const Key_enc = 2                                 'Encoder key      @ PINC.2
Const Key_mode = 3                                'Key mode         @ PINC.3
Const Key_right = 4                               'Key cursor right @ PINC.4
Const Key_left = 5                                'Key cursor left  @ PINC.5

'Configure LCD on Port B and D ------------------------------------------------
'PD5  LCD      RS
'PD6  LCD      E
'PD7  LCD      DB4
'PB0  LCD      DB5
'PB1  LCD      DB6
'PB2  LCD      DB7

Config Lcd = 16 * 2
Config Lcdpin = Pin , Db7 = PortB.2 , Db6 = PortB.1 , Db5 = PortB.0
Config Lcdpin = Pin , Db4 = PortD.7 , E = PortD.6 , Rs = PortD.5
Config Lcdbus = 4
Const bytChar = 16                                'Characters per LCD line

'Variables program control ----------------------------------------------------
Dim bitChanged As Bit                             '=1: Setup, value changed
Dim bytMode As Byte                               'Index of mode
Dim bytCom_Header As Byte                         'Communication header
Dim bytSave As Byte
Dim bytCursorPos As Byte                          'Current cursor position
Dim bytCursorFirst As Byte                        'First LCD cursor position
Dim bytCursorLast As Byte                         'Last LCD cursor position
Dim bytSep As Byte                                'Thousands separator

'Common Variables, multiply used ----------------------------------------------
Dim bytTmp0 As Byte
Dim bytTmp1 As Byte
Dim dwdFreq As DWord
Dim dwdFreq_step As DWord
Dim dwdVal As DWord
Dim sngTmp0 As Single
Dim dblTmp0 As Double
Dim dblTmp1 As Double
Dim dblTmp2 As Double
Dim strTmp1 As String * 1
Dim strTmp9 As String * 9                         'Freg. string, no separators
Dim bytTmp9(9) As Byte At strTmp9 Overlay         'the same as byte array
Dim strDis As String * 16                         'Text to be dispayed on LCD
Dim bytDis(16) As Byte At strDis Overlay          'the same as byte array
Dim strCom14 As String * 14                       'VFO communication message
Dim bytCom14(13) As Byte At strCom14 Overlay      'the same as byte array

'Variables keys ---------------------------------------------------------------
Dim bytIwr0 As Byte
Dim bytKey_ct0 As Byte
Dim bytKey_ct1 As Byte
Dim bytKey_state As Byte
Dim bytKey_press As Byte
Dim bytKey_save As Byte
Dim bytKey_mask As Byte
Dim bitKey_left As Bit                            '=1 if key left is pressed
Dim bitKey_enc As Bit                             '=1 if key encoder is pressed
Dim bytKey_mode As Byte                           'Mode  key pressed: 1(short), 2(long)
Dim bytKey_right As Byte                          'Right key pressed: 1(short), 2(long)
Dim wrdTimer0_runs As Word
Const Key_short = 1                               'Key pressed short
Const Key_long = 2                                'Key pressed long

'Variables Encoder  -----------------------------------------------------------
Dim bytIncr As Byte                               '=2: Encoder incr., =1: decr., =0: ./.
Dim bytEnc_old As Byte
Dim bytEnc_new As Byte
Const DoubleStep = 0                              '=1: intermediate position detected

'EEPROM data ------------------------------------------------------------------
Dim wrdAddr As Word
Dim dwdEEPROM As DWord                            'DWord to/from EEPROM
Dim bytDwdEEP(4) As Byte At dwdEEPROM Overlay
Dim dwdClock As DWord                             'NWT DDS clock frequency
Dim dwdFreq_min As DWord                          'Min. NWT frequency
Dim dwdFreq_max As DWord                          'Max. NWT frequency

'Declare subs & functions -----------------------------------------------------
Declare Sub VFOFreq(ByVal dwdF As DWord , byVal bytHead As Byte , byVal bytSend As Byte)
Declare Sub DDSCal(byVal dwdFreq As DWord)
Declare Sub GetFirstNumber
Declare Sub SetValue
Declare Sub ClearLCDLine(byVal bytRow As Byte , byVal bytChars As Byte)
Declare Function KeyPressedLong(byVal bytKey_Long As Byte) As Byte
Declare Sub CheckEEPROM
Declare Sub EEPROMReadDwrd
Declare Sub EEPROMWriteDwrd
Declare Function Dec2Hex(byVal dblDec As Double , byVal bytLen As Byte) As String

'Configure Timer0-Interrupt for Keys (Timer0 = 8bit, 2^8=256 counts) ----------
'Overflow time: Overflow-Counts * Prescale / crystal frequency
'Possible Prescales: 8, 64, 256, 1024, here 1024
'Timer0 overflow counts are reduced by PresetTimer0 =136 to 256-136=120
'= 120 * 1024 / 14,745,600 = 9,3 msec = 120 Hz (crystal 14.745 MHz)
Config Timer0 = Timer , Prescale = 1024
Const PresetTimer0 = 136                          'Initialize Timer0 (shorten time, >0)
On Timer0 Timer0_isr                              'On Overflow go to Timer0_isr
Const wrdTimer0_long = 70                         'Timer0 counts for long press

'Initialize data ----------------------------------------------------------
bytKey_ct0 = 0                                    'Keys
bytKey_ct1 = 0
Decr bytKey_ct0
bytKey_ct1 = bytKey_ct0                           'Keys continued
bytKey_state = 0
bytKey_press = 0
bytKey_mode = 0                                   'End of key data
bytKey_Mask = &B00011000                          'Mask for short/long keys
'                                                  Long: bits 3 & 4 (right, mode)
bytCom_Header = &H8F                              'NWT PIC message header
bytMode = 11                                      'VFO mode
dwdFreq_step = 1                                  'Start frequency step
bytSep = 44                                       'Thousands separator ASCII ","
bytCursorLast = 11

Cls
Cursor Off
Call ClearLCDLine(1 , bytChar)
LCD "NWT-VFO 1.02"
Call ClearLCDLine(2 , bytChar)
LCD "DL6GL 2022-02-19"
Wait 2
Cls

'Enable interrupts and go -----------------------------------------------------
Enable Timer0
SREG.7 = 1                                        'Enable all interrupts
Timer0 = Presettimer0                             'initialize Timer0

Call CheckEEPROM                                  'Init/Read EEPROM data
dwdFreq = dwdFreq_min                             'Start frequency
Call VFOFreq(dwdFreq , 1 , 1)                     'Show start frequency

Do

   'Check keys, any key pressed? ----------------------------------------------
   'bytKey_press returned from Timer0_isr, flag bits 0...7 = keys 0...7
   'Just pressed detected in Timer0_isr: pressed status flag bit = 1
   'After evaluating key pressed status flag, bit must be reset
   'Short/long pressed status flag is reset in function KeyPressedLong
   'Key action is stored in the variables:
   '-  bitKey_left   =1: left key was pressed,    =0: else
   '-  bitKey_enc    =1: encoder key was pressed, =0: else
   '-  bytKey_mode   =0: mode  key was not pressed
   '-  bytKey_mode   =1: mode  key was pressed short (Key_short=1)
   '-  bytKey_mode   =2: mode  key was pressed long  (Key_long =2)
   '-  bytKey_right  =0: right key was not pressed
   '-  bytKey_right  =1: right key was pressed short (Key_short=1)
   '-  bytKey_right  =2: right key was pressed long  (Key_long =2)

   If bytKey_press > 0 Then                       'Any key was pressed
      bitKey_left = bytKey_press.Key_left         '=1: Left key pressed
      bitKey_enc = bytKey_press.Key_enc           '=1: Encoder key pressed
      bytKey_press.Key_left = 0                   'Reset left key pressed flag
      bytKey_press.Key_enc = 0                    'Reset encoder key pressed flag
      If bytKey_press.Key_mode = 1 Then           '=1: Mode key pressed
         bytTmp0 = KeyPressedLong(Key_mode)       'Pressed short or long?
         Select Case bytTmp0
            Case 1                                'Mode key pressed short
               bytKey_mode = Key_short
            Case 2                                'Mode key pressed long
               bytKey_mode = Key_long
         End Select
      Else
         bytKey_mode = 0                          'Store key not pressed
      End If
      If bytKey_press.Key_right = 1 Then          '=1: Right key pressed
         bytTmp0 = KeyPressedLong(Key_right)      'Pressed short or long?
         Select Case bytTmp0
            Case 1                                'Right key pressed short
               bytKey_right = Key_short
            Case 2                                'Right key pressed long
               bytKey_right = Key_long
         End Select
      Else
         bytKey_right = 0                         'Right key not pressed
      End If
   End If

   'Check mode key ------------------------------------------------------------
   'Mode levels:
   '11   VFO, set frequency with encoder
   '12   VFO, set frequency step for encoder action
   '21   Setup, set NWT DDS clock frequency
   '22   Setup, set min. VFO frequency
   '23   Setup, set max. VFO frequency

   If bytKey_mode = Key_long Then                 'Set key pressed long
      If bytMode < 20 Then
         bytMode = 21                             'Change to setup, 1st item
         Cls
         Locate 1 , 1
         LCD "DDS clock freq. "
         bitChanged = 0
         Call VFOFreq(dwdClock , 0 , 0)
         Call GetFirstNumber
         bytKey_mode = 0                          'Reset mode key flag
         Cursor ON
      Else
         bytMode = 11                             'Change to VFO
      End If
   End If

   'Check Encoder key, modes 11 & 12 ------------------------------------------
   'Mode 11: Set VFO frequency
   'Mode 12: Set VFO frequency step
   'Set default data
   If bytMode < 13 And bitKey_enc = 1 Then        'Encoder key pressed
      Incr bytMode
      Cls
      Select Case bytMode
         Case 11                                  'Set VFO frequency
            Call VFOFreq(dwdFreq , 1 , 1)
         Case 12                                  'Set VFO frequency step
            Locate 1 , 1
            LCD "Set freq. step  "
            Call VFOFreq(dwdFreq_step , 0 , 0)
         Case 13                                  'from mode 12 to 11
            bytMode = 11
            Call VFOFreq(dwdFreq , 1 , 1)
      End Select
      bitKey_Enc = 0                              'Reset encoder key flag
   End If

   'Check mode key, modes 21 to 23 (setup) ------------------------------------
   'Setup levels:
   '21   Setup, set NWT DDS clock frequency
   '22   Setup, set min. VFO frequency
   '23   Setup, set max. VFO frequency

   If bytMode > 19 And bytKey_mode = Key_short Then
      Incr bytMode
      Cls
      Select Case bytMode                         'Prepare data for setup
         Case 22                                  'set min. VFO frequency
            Locate 1 , 1
            LCD "Min. VFO freq.  "
            bitChanged = 0
            Call VFOFreq(dwdFreq_min , 0 , 0)
            Call GetFirstNumber
         Case 23                                  'set max. VFO frequency
            Locate 1 , 1
            LCD "Max. VFO freq.  "
            bitChanged = 0
            Call VFOFreq(dwdFreq_max , 0 , 0)
            Call GetFirstNumber
         Case Else                                'Back to VFO mode 11
            bytMode = 11
            Call VFOFreq(dwdFreq , 1 , 1)         'Show VFO frequency again
            Cursor off NoBlink
      End Select
      bytKey_mode = 0                             'Reset mode key flag
   End If

   'Handle dynamic mode data --------------------------------------------------
   Select Case bytMode

      Case 11                                     'VFO mode, set frequency
         Disable Timer0
         Select Case bytIncr
            Case 2                                'Encoder incremented
               dwdFreq = dwdFreq + dwdFreq_step
               If dwdFreq > dwdFreq_max Then
                  dwdFreq = dwdFreq_max
               End If
            Case 1                                'Encoder decremented
               If dwdFreq > dwdFreq_step Then
                  dwdFreq = dwdFreq - dwdFreq_step
               Else
                  dwdFreq = dwdFreq_min
               End If
         End Select
         If bytIncr > 0 Then                      'Show new frequency
            Call VFOFreq(dwdFreq , 0 , 1)
            bytIncr = 0
         End If
         Enable TIMER0

      Case 12                                     'VFO mode, set freq. step
         Disable Timer0
         Select Case bytIncr
            Case 2                                'Encoder incremented
               dwdFreq_step = dwdFreq_step * 10
               If dwdFreq_step > 10000000 Then    '> 10 MHz
                  dwdFreq_step = 10000000
               End If
            Case 1                                'Encoder decremented
               If dwdFreq_step > 9 Then           '> 9 Hz
                  dwdFreq_step = dwdFreq_step / 10
               End If
         End Select
         If bytIncr > 0 Then                      'Show new frequency step
            Call VFOFreq(dwdFreq_step , 0 , 0)
            bytIncr = 0
         End If
         Enable TIMER0

      Case 21                                     'Set NWT DDS clock frequency
         Call SetValue
         If bitChanged = 1 Then
            dwdClock = dwdVal
         End If

      Case 22                                     'Set min. VFO frequency
         Call SetValue
         If bitChanged = 1 Then
            dwdFreq_min = dwdVal
         End If

      Case 23                                     'Set max. VFO frequency
         Call SetValue
         If bitChanged = 1 Then
            dwdFreq_max = dwdVal
         End If

   End Select

Loop

End

'==============================================================================
Timer0_isr:
'Encoder interrupt on Timer0 overflow
'While data are displayed Timer0 should be stopped
'Output: bytIncr = 2: increment, =1: decrement if encoder was changed
'Having processed the bytIncr data in the main program reset bytIncr = 0
'Switching pattern of ALPS STEC11 encoder:
'Turning clockwise              ... counter clockwise
'A    B                             A     B
'1    1  start detent position      1     1
'1    0  intermediate position 1    0     1
'0    0  intermediate position 2    0     0
'0    1  intermediate position 3    1     0
'1    1  next detent position       1     1

   Timer0 = PresetTimer0

   'Encoder -------------------------------------------------------------------
   bytEnc_new.0 = bitEnc_A                        'Set bit 0 of bytEnc_new
   bytEnc_new.1 = bitEnc_B                        'Set bit 1 of bytEnc_new

   'Check encoder pattern if encoder is changed
   'bytIncr = 1: decremented (turned counter clockwise)
   'bytIncr = 2: incremented (turned         clockwise)
   If bytEnc_new <> bytEnc_old Then               'encoder is changed
      'from intermediate position 3 to quiescent position
      If bytEnc_new = &B00000011 And bytEnc_old = &B00000010 Then bytIncr = 2
      If bytEnc_new = &B00000011 And bytEnc_old = &B00000001 Then bytIncr = 1
      'from intermediate position 1 to intermediate position 2
      'DoubleStep = 1:  intermediate position 2 gives an additional tick
      '                 2 ticks between 2 detent positions on slow rotation
      '           = 0:  1 tick  between 2 detent positions
      #If DoubleStep = 1
         If bytEnc_new = &B00000000 And bytEnc_old = &B00000001 Then bytIncr = 2
         If bytEnc_new = &B00000000 And bytEnc_old = &B00000010 Then bytIncr = 1
      #EndIf
      bytEnc_old = bytEnc_new                     'old <- new for the next time
   End If

   'Keys ----------------------------------------------------------------------
   'got from Thomas Baer in microcontroller.net
   'Original (ASM) by Peter Dannegger
   'http://www.mikrocontroller.net/topic/tasten-entprellen-bulletproof

   'Input:  Key_port       Each of the 8 bits of byte variable represents one key
   '                       bits 0...7: keys 0...7 @ Key_port
   '        bytKey_mask    Bit pattern to mask the keys enabled for short/long
   '                       Key number (0...7) to be checked for short/long
   '                       e.g. "00000101" Keys 0 and 2 are short/long keys
   'Output: bytKey_press   bit pattern, key which changed from open to pressed
   '                       Reset after evaluation in the main program
   '                       Bits remain = 1 until resetted in the main program
   '        bytKey_state   bit pattern, Key state, 1:pressed or 0:not pressed
   '                       Bits = 1 while key is pressed, = 0: Key is released
   '        bytKey_save    last state of bytKey_state before entering Timer0_isr
   '                       So it can be determined wether between two calls
   '                       of Timer0_isr any key was pressed.

   bytKey_save = bytKey_state                     'save old status before keys were pressed
   bytIwr0 = Key_port                             'get Bits of the KeyPort
   bytIwr0 = Not bytIwr0
   bytIwr0 = bytIwr0 Xor bytKey_state
   bytKey_ct0 = bytKey_ct0 And bytIwr0
   bytKey_ct1 = bytKey_ct1 And bytIwr0
   bytKey_ct0 = Not bytKey_ct0
   bytKey_ct1 = bytKey_ct1 Xor bytKey_ct0
   bytIwr0 = bytIwr0 And bytKey_ct0
   bytIwr0 = bytIwr0 And bytKey_ct1
   bytKey_state = bytKey_state Xor bytIwr0
   bytIwr0 = bytIwr0 And bytKey_state
   bytKey_press = bytKey_press Or bytIwr0

   'Counter for short/long detection
   bytIwr0 = bytKey_state AND bytKey_mask         'Mask out short/long keys
   If bytIwr0 > 0 Then                            'Short/long key still pressed
      Incr wrdTimer0_runs                         'Clock is running...
   End If

Return

'==============================================================================
Sub SetValue
'Set EEPROM value via keys and restore it to EEPROM
'Digits are selected via right/left keys
'Numbers are changed via encoder (0...9)
'Input:  bytRow         LCD row to display value
'        stsDis         Formatted freq. string "999,999,999"
'        bytCursorPos            Current cursor start position

'        bytType                 Type of value
'        wrdAddr                 EEPROM address to store value

   'Select digit to the right (Key_Right)
   'Key_Right pressed short: select digit
   'Key_Right pressed long : store value to EEPROM

   If bytKey_right <> 0 Then                      'Key_Right pressed
      If bytKey_right = Key_short Then            'Pressed short
         If bytCursorPos < bytCursorLast Then
            Incr bytCursorPos
            If bytCursorPos = 4 Or bytCursorPos = 8 Then       'Is thousands separator
               Incr bytCursorPos                  'skip it
            End If
         End If
      Else                                        'Pressed long: store
         Select Case bytMode
            Case 21                               'Store NWT DDS clock freq
               wrdAddr = 2
               dwdEEPROM = dwdClock
               Call EEPROMWriteDwrd               'Write NWT DDS clock frequency
               Call DDSCal(dwdClock)              'Calculate calib. hex message
               For bytTmp0 = 1 To 14              'and send it to NWT
                  bytTmp1 = bytCom14(bytTmp0)
                  PrintBin bytTmp1
               Next bytTmp0
            Case 22                               'Store min. VFO frequency
               wrdAddr = 6
               dwdEEPROM = dwdFreq_min
               Call EEPROMWriteDwrd               'Write NWT min. frequency
            Case 23                               'Store max. VFO frequency
               wrdAddr = 10
               dwdEEPROM = dwdFreq_max
               Call EEPROMWriteDwrd               'Write NWT max. frequency
         End Select
         Locate 2 , 16
         LCD "S"
      End If
      bytKey_right = 0                            'Reset right Key flag
   End If

   'Select digit downwards (Key_Left)
   If bitKey_left = 1 Then                        'Key_Left pressed
      If bytCursorPos > 1 Then
         Decr bytCursorPos
         If bytCursorPos = 4 Or bytCursorPos = 8 Then       'Is thousands separator
            Decr bytCursorPos                     'skip it
         End If
      End If
      bitKey_left = 0                             'Reset left Key flag
   End If

   'Change selected digit via Encoder
   Locate 2 , bytCursorPos
   Select Case bytIncr
      Disable Timer0
      bytSave = bytDis(bytCursorPos)              'Char (ASCII) to be changed
      Case 1                                      'Encoder decremented
         If bytSave <= 48 Then                    '<= ASCII 0
            bytSave = 57                          'ASCII 9
         Else                                     '> ASCII 0
            Decr bytSave
         End If
      Case 2                                      'Encoder incremented
         If bytSave >= 57 Then                    '>= ASCII 9
            bytSave = 48                          'ASCII 0
         Else                                     '< ASCII 9
            Incr bytSave
         End If
   End Select
   If bytIncr > 0 Then                            'Show changed digit
      bytIncr = 0
      bitChanged = 1                              'Value changed
      strTmp1 = Chr(bytSave)                      'Changed value (ASCII bytSave)
      bytDis(bytCursorPos) = strTmp1              '... into string strDis
      LCD strTmp1                                 'Show changed digit
      bytTmp1 = 0
      strTmp9 = Space(9)
      For bytTmp0 = 1 To bytCursorLast            'Recover DWord number
         If bytDis(bytTmp0) > 47 And bytDis(bytTmp0) < 58 Then
            Incr bytTmp1
            bytTmp9(bytTmp1) = bytDis(bytTmp0)
         End If
      Next bytTmp0
      dwdVal = Val(strTmp9)                       'Changed DWord number
   End If
   Enable Timer0

End Sub

'==============================================================================
Sub GetFirstNumber
'Get first number <> Space bytCursorFirst in frequency string strDis

   For bytTmp0 = 1 To bytCursorLast               'Find first character <> Space
      If bytDis(bytTmp0) <> 32 Then               'Not equal space
         bytCursorFirst = bytTmp0
         Exit For
      End If
   Next bytTmp0
   bytCursorPos = bytCursorFirst

End Sub

'==============================================================================
Sub VFOFreq(ByVal dwdF As DWord , byVal bytHead As Byte , byVal bytSend As Byte)
'Show VFO frequency dwdF and send message to NWT for VFO setting
'Communication message string:
'Byte  1:         bytCom_Header = 8F hex
'Byte  2:         "f"
'Bytes 3 to 11:   frequency (Hz), 9 bytes (digits) with leading zeros
'Input:  dwdF           Frequency (Hz)
'        bytHead        LCD option
'                       = 0: no header in row 1
'                       = 1: header in row 1
'        bytSend        NWT send option
'                       = 0: do not send to NWT
'                       = 1: send frequency message to NWT
'        bytCom_Header  Protocol header to NWT PIC

   'Optional header in LCD row 1
   If bytHead = 1 Then                            'Show header
      Locate 1 , 1
      LCD "VFO frequency   "
   End If

   'Show frequency on LCD row 2
   strTmp1 = Chr(bytSep)                          'Thousands separator
   strTmp9 = Str(dwdF)
   strTmp9 = Format(strTmp9 , "        0")
   strDis = Left(strTmp9 , 3) + strTmp1 + Mid(strTmp9 , 4 , 3) + strTmp1 + Right(strTmp9 , 3)

   If Left(strTmp9 , 3) = "   " Then              'Cancel leading separators ","
      For bytTmp0 = 1 To 4
         bytDis(bytTmp0) = " "
      Next bytTmp0
   End If
   If Mid(strTmp9 , 4 , 3) = "   " Then
      For bytTmp0 = 5 To 8
         bytDis(bytTmp0) = " "
      Next bytTmp0
   End If

   Locate 2 , 1
   Lcd strDis ; " Hz"

   'Send message to NWT
   If bytSend = 1 Then
      strCom14 = str(dwdF)                        'Make message string...
      strCom14 = Format(strCom14 , "  000000000")
      bytCom14(1) = bytCom_Header
      bytCom14(2) = "f"
      For bytTmp0 = 1 To 11                       '11 bytes VFO message
         bytTmp1 = bytCom14(bytTmp0)
         PrintBin bytTmp1                         '... and send it
      Next bytTmp0
   End If

End Sub

'==============================================================================
Sub DDSCal(byVal dwdFreq As DWord)
'Calculate DDS calibration string
'Calibration constant: 2^64 / DDS-frequency (Hz)
'Communication message string:
'Byte   1:        bytCom_Header = 8F hex
'Byte   2:       "e"
'Bytes  3 to 12:  Calibration constant, 10 bytes hex with leading zeros
'Bytes 13 to 14:  "00"
'Input:  dwdFreq  DDS-frequency (round about 400 MHz)

   dblTmp0 = dwdFreq                              'Convert to double
   dblTmp1 = 4294967296                           '2^32 = 4,294,967,296
   dblTmp2 = dblTmp1 / dblTmp0                    '=2^32 / freq.
   dblTmp2 = dblTmp2 * dblTmp1                    '=2^64 / freq.
   dblTmp2 = Int(dblTmp2)                         'Integer for hex conversion
   strDis = Dec2Hex(dblTmp2 , 9)                  'Convert constant to hex

   bytCom14(1) = bytCom_Header                    'Make message string ...
   bytCom14(2) = "e"
   bytCom14(3) = "0"
   bytTmp1 = 3
   For bytTmp0 = 1 To 9
      Incr bytTmp1
      bytCom14(bytTmp1) = bytDis(bytTmp0)
   Next bytTmp0
   Incr bytTmp1
   bytCom14(bytTmp1) = "0"
   Incr bytTmp1
   bytCom14(bytTmp1) = "0"

   'Show calibration message string
   'Call ClearLCDLine(2 , 16)
   'LCD strCom14

End Sub

'==============================================================================
Function KeyPressedLong(byVal bytKey_Long As Byte) As Byte
   'Check if key bytKey_Long (0...7) in Key_port is pressed short or long
   'Output: KeyPressedLong =  0: evaluating, key is still pressed
   '                          1: key is released, pressed short
   '                          2: key is released, pressed long

   If bytKey_state.bytKey_Long = 1 Then           'Key still pressed
      KeyPressedLong = 0
   Else                                           'Key released
      If wrdTimer0_runs < wrdTimer0_long Then     'short time elapsed
         KeyPressedLong = 1                       'Pressed short
      Else
         KeyPressedLong = 2                       'Pressed short
      End If
      wrdTimer0_runs = 0
      bytKey_press.bytKey_Long = 0                'Reset pressed flag
   End If

End Function

'==============================================================================
Sub ClearLCDLine(byVal bytRow As Byte , byVal bytChars As Byte)
'Clear LCD bytChars characters in line bytRow
   Locate bytRow , 1
   LCD SPC(bytChars)
   Locate bytRow , 1

End Sub

'==============================================================================
Sub CheckEEPROM
'Check if NWT data are stored in EEPROM
'if not, write NWT data
'Read EEPROM data
'In a new controller EEPROM data = 255
'NWT data:
'Byte Variable    Type     Description
' 0   --          Byte     Not used
' 1   --          Byte     = 1: data are stored
' 2   dwdClock    DWord    NWT DDS clock frequency
' 6   dwdFreq_min DWord    NWT min. frequency
'10   dwdFreq_max DWord    NWT max. frequency

   ReadEeprom bytTmp1 , 1                         'Check Address #1

   If bytTmp1 = 1 Then                            'Read available EEPROM data
      Call ClearLCDLine(1 , bytChar)
      LCD "Read EEPROM..."
      wrdAddr = 2
      Call EEPROMReadDwrd                         'Read NWT DDS clock frequency
      dwdClock = dwdEEPROM
      Call EEPROMReadDwrd                         'Read NWT min. frequency
      dwdFreq_min = dwdEEPROM
      Call EEPROMReadDwrd                         'Read NWT max. frequency
      dwdFreq_max = dwdEEPROM
      Call ClearLCDLine(2 , bytChar)
      Lcd "... Read done"
      Wait 1
   Else                                           'Is a brandnew controller
      Call ClearLCDLine(1 , bytChar)              'Initialize it with data
      LCD "Init EEPROM..."
      wrdAddr = 2
      'dwdClock = 400000000                        'Clock freq. default (400 MHz)
      dwdClock = 399997120                        'Clock freq. 2002-02-20
      dwdEEPROM = dwdClock
      Call EEPROMWriteDwrd                        'Write NWT DDS clock frequency
      dwdFreq_min = 10                            'NWT min. frequency
      dwdEEPROM = dwdFreq_min
      Call EEPROMWriteDwrd                        'Write NWT min. frequency
      dwdFreq_max = 160000000                     'NWT max. frequency
      dwdEEPROM = dwdFreq_max
      Call EEPROMWriteDwrd                        'Write NWT max. frequency

      bytTmp0 = 1
      wrdAddr = 1
      WriteEEPROM bytTmp0 , wrdAddr
      Call ClearLCDLine(2 , bytChar)
      Lcd "... Write done"
      Wait 1
   End If

End Sub


'==============================================================================
Sub EEPROMWriteDwrd
   'Decode dwdDWord to 4 bytes of bytDwdEEP for writing to EEPROM
   'Input dwdEEPROM is overlayed with bytDwdEEP(i), i=1...4
   'The 4 bytes bytDwdEEP(i)are written to EEPROM
   'Input:  dwdEEPROM to be decoded & written to EEPROM
   '        wrdAddr first EEPROM address

   For bytTmp0 = 1 To 4
      WriteEEPROM bytDwdEEP(bytTmp0) , wrdAddr
      Incr wrdAddr
   Next bytTmp0

End Sub

'==============================================================================
Sub EEPROMReadDwrd
   'Encode 4 bytes read from EEPROM to sngSingle
   'The 4 bytes are read to bytDwdEEP(i), i=1...4
   'bytDwdEEP is overlayed @ dwdEEPROM
   'Input:  wrdAddr first EEPROM address
   'Output: dwdEEPROM encoded to single variable read from EEPROM


   For bytTmp0 = 1 To 4
      ReadEEPROM bytDwdEEP(bytTmp0) , wrdAddr
      Incr wrdAddr
   Next bytTmp0

End Sub

'==============================================================================
Function Dec2Hex(byVal dblDec As Double , byVal bytLen As Byte) As String
'Convert decimal dblDec to hex
'Input:  dblDec   decimal number
'        bytLen   Length of Hex string
'Output: Dec2Hex  Hex string, bytLen characters long
'Scratch string strTmp9 must be declared in the main program
'bytTmp9(1...bytLen) must be overlayed with strTmp9

   bytTmp0 = bytLen + 1                           'bytTmp9 byte counter

   Do

      dblTmp0 = dblDec / 16
      dblTmp0 = Int(dblTmp0)
      dblTmp1 = dblTmp0 * 16
      dblTmp1 = dblDec - dblTmp1
      sngTmp0 = dblTmp1                           'Convert to single...
      bytTmp1 = Int(sngTmp0)                      '...does not work with double

      Select Case bytTmp1                         'Get hex number
         Case Is < 10
            strTmp1 = Str(bytTmp1)
         Case 10
            strTmp1 = "A"
         Case 11
            strTmp1 = "B"
         Case 12
            strTmp1 = "C"
         Case 13
            strTmp1 = "D"
         Case 14
            strTmp1 = "E"
         Case 15
            strTmp1 = "F"
      End Select
      Decr bytTmp0
      bytTmp9(bytTmp0) = strTmp1                  'Fill bytes backwards
      dblDec = dblTmp0
   Loop Until dblDec < 1

   Dec2Hex = strTmp9

End Function